@@ -1,14 +1,16 @@
+#define _GNU_SOURCE 1
 #include <stdio.h>
 #include <stdlib.h>
 #include <fcntl.h>
 #include <errno.h>
-#ifndef WIN32
+#ifndef _WIN32
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <signal.h>
+#include <poll.h>
 #else
 #include <windows.h>
 #include <sddl.h>
@@ -22,11 +24,11 @@
 #include <vchan-common.h>
 #endif
 
-#ifndef WIN32
+#ifndef _WIN32
 #ifdef HAVE_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
-#else // !WIN32
+#else // !_WIN32
 // parameters for a client pipe thread
 struct thread_param {
     struct db_daemon_data *daemon;
@@ -40,7 +42,7 @@ struct thread_param {
 
 int init_vchan(struct db_daemon_data *d);
 
-#ifndef WIN32
+#ifndef _WIN32
 int sigterm_received = 0;
 static void sigterm_handler(int s) {
     sigterm_received = 1;
@@ -123,7 +125,7 @@ static int accept_new_client(struct db_daemon_data *d) {
     return add_client(d, new_client_fd);
 }
 
-#else // !WIN32
+#else // !_WIN32
 
 /* Main pipe server processing loop (separate thread).
  * Takes care of accepting clients and receiving data.
@@ -371,50 +373,47 @@ void close_server_socket(struct db_daemon_data *d) {
         QpsDestroy(d->pipe_server);
     d->pipe_server = NULL;
 }
-#endif // WIN32
+#endif // _WIN32
 
-#ifndef WIN32
+#ifndef _WIN32
 
-static int fill_fdsets_for_select(struct db_daemon_data *d,
-        fd_set * read_fdset, fd_set * write_fdset) {
+static size_t fill_fdsets_for_select(struct db_daemon_data *d,
+        struct pollfd fds[static MAX_CLIENTS + 2]) {
     struct client *client;
-    int max_fd;
+    size_t total_fds = 2;
 
-    FD_ZERO(read_fdset);
-    FD_ZERO(write_fdset);
-    QUBES_FD_SET(d->socket_fd, read_fdset);
-    max_fd = d->socket_fd;
-    if (d->vchan) {
-        int vchan_fd = libvchan_fd_for_select(d->vchan);
-        QUBES_FD_SET(vchan_fd, read_fdset);
-        if (vchan_fd > max_fd)
-            max_fd = vchan_fd;
-    }
+    fds[0] = (struct pollfd) {
+        .fd = d->socket_fd,
+        .events = POLLIN | POLLHUP,
+        .revents = 0,
+    };
+    fds[1] = (struct pollfd) {
+        .fd = d->vchan ? libvchan_fd_for_select(d->vchan) : -1,
+        .events = POLLIN | POLLHUP,
+        .revents = 0,
+    };
 
     client = d->client_list;
     while (client) {
+        assert(total_fds < MAX_CLIENTS + 2);
         /* Do not read commands from client, which have some buffered data,
          * first try to send them all. If client do not handle write buffering
          * properly, it can cause a deadlock there, but at least qubesdb-daemon
          * will still handle other requests */
-        assert(client->fd < FD_SETSIZE);
-        if (buffer_datacount(client->write_queue))
-            QUBES_FD_SET(client->fd, write_fdset);
-        else
-            QUBES_FD_SET(client->fd, read_fdset);
-        if (client->fd > max_fd)
-            max_fd = client->fd;
+        fds[total_fds++] = (struct pollfd) {
+            .fd = client->fd,
+            .events = buffer_datacount(client->write_queue) ? POLLOUT : POLLIN | POLLHUP,
+            .revents = 0,
+        };
         client = client->next;
     }
-    return max_fd;
+    return total_fds;
 }
 
 static int mainloop(struct db_daemon_data *d) {
-    fd_set read_fdset;
-    fd_set write_fdset;
     struct client *client;
-    int max_fd;
     int ret;
+    static struct pollfd fds[MAX_CLIENTS + 2];
     sigset_t sigterm_mask;
     sigset_t oldmask;
     struct timespec ts = { 10, 0 };
@@ -423,7 +422,10 @@ static int mainloop(struct db_daemon_data *d) {
     sigaddset(&sigterm_mask, SIGTERM);
 
     while (1) {
-        max_fd = fill_fdsets_for_select(d, &read_fdset, &write_fdset);
+        size_t current_fd = 2;
+        size_t const nfds = fill_fdsets_for_select(d, fds);
+        assert(nfds >= 2);
+        assert(nfds <= MAX_CLIENTS + 2);
 
         if (sigprocmask(SIG_BLOCK, &sigterm_mask, &oldmask) < 0) {
             perror("sigprocmask");
@@ -433,23 +435,18 @@ static int mainloop(struct db_daemon_data *d) {
             fprintf(stderr, "terminating\n");
             break;
         }
-        ret = pselect(max_fd+1, &read_fdset, &write_fdset, NULL, &ts, &oldmask);
+        ret = ppoll(fds, nfds, &ts, &oldmask);
         if (ret < 0) {
             if (errno == EINTR)
                 continue;
-            /* client could have disconnected just before select call, so
-             * ignore this error and retry
-             * FIXME: This probably will loop indefinitelly */
-            if (errno == EBADF)
-                continue;
-            perror("select");
+            perror("ppoll");
             break;
         }
         /* restore signal mask */
         sigprocmask(SIG_SETMASK, &oldmask, NULL);
 
         if (d->vchan) {
-            if (FD_ISSET(libvchan_fd_for_select(d->vchan), &read_fdset))
+            if (fds[1].revents)
                 libvchan_wait(d->vchan);
             if (!libvchan_is_open(d->vchan)) {
                 fprintf(stderr, "vchan closed\n");
@@ -489,11 +486,13 @@ static int mainloop(struct db_daemon_data *d) {
 
         client = d->client_list;
         while (client) {
-            if (FD_ISSET(client->fd, &write_fdset)) {
+            assert(current_fd < MAX_CLIENTS + 2);
+            short revents = fds[current_fd++].revents;
+            if (revents & POLLOUT) {
                 /* just send bufferred data, possibly not all of them */
                 write_client_buffered(client, NULL, 0);
             }
-            if (FD_ISSET(client->fd, &read_fdset)) {
+            if (revents & (POLLIN | POLLHUP)) {
                 if (!handle_client_data(d, client, NULL, 0)) {
                     struct client *client_to_remove = client;
                     client = client->next;
@@ -503,56 +502,65 @@ static int mainloop(struct db_daemon_data *d) {
             }
             client = client->next;
         }
+        assert(current_fd == nfds);
 
-        if (FD_ISSET(d->socket_fd, &read_fdset)) {
+        if (fds[0].revents) {
             accept_new_client(d);
         }
     }
     return 1;
 }
 
-#define MAX_FILE_PATH 256
 static int init_server_socket(struct db_daemon_data *d) {
-    char socket_address[MAX_FILE_PATH];
     struct sockaddr_un sockname;
     int s;
     struct stat stat_buf;
     mode_t old_umask;
 
+    memset(&sockname, 0, sizeof(sockname));
+    sockname.sun_family = AF_UNIX;
+    if (mkdir("/var/run/qubes", 0775) && errno != EEXIST) {
+        perror("mkdir /var/run/qubes");
+        return 0;
+    }
     if (d->remote_name) {
-        snprintf(socket_address, MAX_FILE_PATH,
-                QDB_DAEMON_PATH_PATTERN, d->remote_name);
+        if ((unsigned)snprintf(sockname.sun_path, sizeof sockname.sun_path,
+                               QDB_DAEMON_PATH_PATTERN, d->remote_name) >=
+            sizeof sockname.sun_path) {
+            perror("snprintf()");
+            return 0;
+        }
         if (d->remote_domid == 0) {
             /* the same daemon as both VM and Admin parts */
             unlink(QDB_DAEMON_LOCAL_PATH);
-            if (symlink(socket_address, QDB_DAEMON_LOCAL_PATH) < 0) {
+            if (symlink(sockname.sun_path, QDB_DAEMON_LOCAL_PATH) < 0) {
                 perror("symlink " QDB_DAEMON_LOCAL_PATH);
                 return 0;
             }
         }
     } else {
-        snprintf(socket_address, MAX_FILE_PATH,
-                QDB_DAEMON_LOCAL_PATH);
+        _Static_assert(sizeof QDB_DAEMON_LOCAL_PATH <= sizeof sockname.sun_path,
+                       QDB_DAEMON_LOCAL_PATH "too long");
+        strcpy(sockname.sun_path, QDB_DAEMON_LOCAL_PATH);
     }
 
-    unlink(socket_address);
+    if (unlink(sockname.sun_path) && errno != ENOENT) {
+        perror("unlink() failed");
+        return 0;
+    }
 
     /* make socket available for anyone */
     old_umask = umask(0);
 
     s = socket(AF_UNIX, SOCK_STREAM, 0);
-    memset(&sockname, 0, sizeof(sockname));
-    sockname.sun_family = AF_UNIX;
-    memcpy(sockname.sun_path, socket_address, strlen(socket_address));
-
     if (bind(s, (struct sockaddr *) &sockname, sizeof(sockname)) == -1) {
-        printf("bind() failed\n");
+        perror("bind() failed");
         close(s);
         return 0;
     }
 //      chmod(sockname.sun_path, 0666);
     if (listen(s, SERVER_SOCKET_BACKLOG) == -1) {
-        perror("listen() failed\n");
+        perror("listen() failed");
         close(s);
         return 0;
     }
@@ -563,7 +571,7 @@ static int init_server_socket(struct db_daemon_data *d) {
     return 1;
 }
 
-#endif /* !WIN32 */
+#endif /* !_WIN32 */
 
 int init_vchan(struct db_daemon_data *d) {
     if (d->vchan) {
@@ -585,7 +593,7 @@ int init_vchan(struct db_daemon_data *d) {
             d->vchan = NULL;
             return 1;
         }
-#ifndef WIN32
+#ifndef _WIN32
         d->vchan = libvchan_server_init(d->remote_domid, QUBESDB_VCHAN_PORT, 4096, 4096);
 #else
         // We give a 5 minute timeout here because xeniface can take some time
@@ -597,7 +605,7 @@ int init_vchan(struct db_daemon_data *d) {
         d->remote_connected = 0;
     } else {
         /* VM part: connect to admin domain */
-#ifndef WIN32
+#ifndef _WIN32
         d->vchan = libvchan_client_init(d->remote_domid, QUBESDB_VCHAN_PORT);
 #else
         // We give a 5 minute timeout here because xeniface can take some time
@@ -611,7 +619,7 @@ int init_vchan(struct db_daemon_data *d) {
     return 1;
 }
 
-#ifndef WIN32
+#ifndef _WIN32
 static int create_pidfile(struct db_daemon_data *d) {
     char pidfile_name[256];
     FILE *pidfile;
@@ -675,14 +683,14 @@ static void close_server_socket(struct db_daemon_data *d) {
             unlink(sockname.sun_path);
     }
 }
-#endif // !WIN32
+#endif // !_WIN32
 
 static void usage(char *argv0) {
     fprintf(stderr, "Usage: %s <remote-domid> [<remote-name>]\n", argv0);
     fprintf(stderr, "       Give <remote-name> only in dom0\n");
 }
 
-#ifdef WIN32
+#ifdef _WIN32
 DWORD WINAPI service_thread(PVOID param) {
     PSERVICE_WORKER_CONTEXT ctx = param;
     struct db_daemon_data *d = ctx->UserContext;
@@ -708,7 +716,7 @@ int main(int argc, char **argv) {
 int fuzz_main(int argc, char **argv) {
 #endif
     struct db_daemon_data d;
-#ifndef WIN32
+#ifndef _WIN32
     int ready_pipe[2] = {0, 0};
 #endif
     int ret;
@@ -718,7 +726,7 @@ int fuzz_main(int argc, char **argv) {
         exit(1);
     }
 
-#ifndef WIN32
+#ifndef _WIN32
     memset(&d, 0, sizeof(d));
 #else
     RtlSecureZeroMemory(&d, sizeof(d));
@@ -733,7 +741,7 @@ int fuzz_main(int argc, char **argv) {
     /* if not running under SystemD, fork and use pipe() to notify parent about
      * sucessful start */
     /* FIXME: OS dependent code */
-#ifndef WIN32
+#ifndef _WIN32
 #ifdef HAVE_SYSTEMD
     if (!getenv("NOTIFY_SOCKET")) {
 #else
@@ -754,7 +762,7 @@ int fuzz_main(int argc, char **argv) {
                 exit(1);
             case 0:
                 close(ready_pipe[0]);
-                snprintf(log_path, sizeof(log_path), "/var/log/qubes/qubesdb.%s.log", d.remote_name);
+                snprintf(log_path, sizeof(log_path), "/var/log/qubes/qubesdb.%s.log", d.remote_name ? d.remote_name : "dom0");
 
                 close(0);
                 old_umask = umask(0);
@@ -785,10 +793,10 @@ int fuzz_main(int argc, char **argv) {
     signal(SIGTERM, sigterm_handler);
 #endif
 
-#ifndef WIN32
+#ifndef _WIN32
     d.db = qubesdb_init(write_client_buffered);
 #else
-    libvchan_register_logger(vchan_logger);
+    libvchan_register_logger(vchan_logger, LogGetLevel());
     d.db = qubesdb_init(send_watch_notify);
     InitializeSRWLock(&d.lock);
 #endif
@@ -802,7 +810,7 @@ int fuzz_main(int argc, char **argv) {
         exit(1);
     }
 
-#ifdef WIN32
+#ifdef _WIN32
     d.db->pipe_server = d.pipe_server;
     /* For Windows, vchan is initialized later, after the service starts
        and reports to the OS. Otherwise it can time-out after the first
@@ -817,7 +825,7 @@ int fuzz_main(int argc, char **argv) {
                       NULL, // notification handler
                       NULL // notification context
                       );
-#else /* WIN32 */
+#else /* _WIN32 */
     if (!init_vchan(&d)) {
         fprintf(stderr, "FATAL: vchan initialization failed\n");
         exit(1);
@@ -855,7 +863,7 @@ int fuzz_main(int argc, char **argv) {
     create_pidfile(&d);
 
     ret = !mainloop(&d);
-#endif /* !WIN32 */
+#endif /* !_WIN32 */
 
     close_server_socket(&d);
 
